#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <glm/glm.hpp>
using namespace glm;

#include "render_vert.glsl.c"
#include "render_frag.glsl.c"

#define MAXCOUNT 512

vec2 cMul(vec2 a, vec2 b)
{
  return vec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

void debug_program(GLuint program) {
  GLint status = 0;
  glGetProgramiv(program, GL_LINK_STATUS, &status);
  GLint length = 0;
  glGetProgramiv(program, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetProgramInfoLog(program, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    fprintf(stderr, "program link info:\n%s", info ? info : "(no info log)");
  }
  if (info) {
    free(info);
  }
}

void debug_shader(GLuint shader, GLenum type, const char *source) {
  GLint status = 0;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
  GLint length = 0;
  glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetShaderInfoLog(shader, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    const char *type_str = "unknown";
    switch (type) {
      case GL_VERTEX_SHADER: type_str = "vertex"; break;
      case GL_FRAGMENT_SHADER: type_str = "fragment"; break;
      case GL_COMPUTE_SHADER: type_str = "compute"; break;
    }
    fprintf(stderr, "%s shader compile info:\n%s\nshader source:\n%s", type_str, info ? info : "(no info log)", source ? source : "(no source)");
  }
  if (info) {
    free(info);
  }
}

GLuint vertex_fragment_shader(const char *vert, const char *frag) {
  GLuint program = glCreateProgram();
  {
    GLuint shader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(shader, 1, &vert, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_VERTEX_SHADER, vert);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  {
    GLuint shader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(shader, 1, &frag, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_FRAGMENT_SHADER, frag);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  glLinkProgram(program);
  debug_program(program);
  return program;
}

char *read_file(const char *name) {
  FILE *f = fopen(name, "r");
  fseek(f, 0, SEEK_END);
  long len = ftell(f);
  fseek(f, 0, SEEK_SET);
  char *str = (char *) malloc(len + 1);
  fread(str, len, 1, f);
  str[len] = 0;
  fclose(f);
  return str;
}

struct {
  // uniform state
  vec2 center;
  vec2 radius;
  vec2 aspect;
  bool julia;
  int  count;
  vec2 inflection[MAXCOUNT];
  // uniform location
  GLint u_center;
  GLint u_radius;
  GLint u_aspect;
  GLint u_julia;
  GLint u_count;
  GLint u_inflection;
  // attrib location
  GLint a_point;
  // other stuff
  int width;
  int height;
  int max_count;
} S;

void update_inflections()
{
  if (S.count < 0)
    return;
  if (S.count > MAXCOUNT)
    return;
  if (S.count > S.max_count)
    S.max_count = S.count;
  glUniform1i(S.u_count, S.count);
  if (S.count > 0)
    glUniform2fv(S.u_inflection, S.count, &S.inflection[MAXCOUNT - S.count][0]);
}

void add_point(vec2 point)
{
  if (S.count >= MAXCOUNT)
    return;
  S.count++;
  S.inflection[MAXCOUNT - S.count] = point;
  update_inflections();
}

void undo_add_point()
{
  if (S.count <= 0)
    return;
  S.count--;
  update_inflections();
}

void redo_add_point()
{
  if (S.count >= S.max_count)
    return;
  S.count++;
  update_inflections();
}

void keycb(GLFWwindow *window, int key, int scancode, int action, int mods)
{
  (void) scancode;
  if (action == GLFW_PRESS)
    switch (key)
    {
      case GLFW_KEY_Q:
      case GLFW_KEY_ESCAPE:
        glfwSetWindowShouldClose(window, GL_TRUE);
        break;
      case GLFW_KEY_HOME:
        S.center = vec2(0.0f, 0.0f);
        S.radius = vec2(2.0f, 0.0f);
        glUniform2fv(S.u_center, 1, &S.center[0]);
        glUniform2fv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_UP:
        S.radius *= (mods & GLFW_MOD_SHIFT) ? 0.1f : 0.5f;
        glUniform2fv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_DOWN:
        S.radius *= (mods & GLFW_MOD_SHIFT) ? 10.0f : 2.0f;
        glUniform2fv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_J:
        S.julia = true;
        glUniform1i(S.u_julia, 1);
        break;
      case GLFW_KEY_M:
        S.julia = false;
        glUniform1i(S.u_julia, 0);
        break;
      case GLFW_KEY_0:
        S.count = 0;
        update_inflections();
        break;
      case GLFW_KEY_MINUS:
        undo_add_point();
        break;
      case GLFW_KEY_EQUAL:
        redo_add_point();
        break;
      case GLFW_KEY_H:
        fprintf(stderr,
          "inflector-gadget 0.1 (GPL) 2017-02-11 Claude Heiland-Allen\n"
          "keyboard controls:\n"
          "\tESC, Q\tquit\n"
          "\tHOME\treset view\n"
          "\tUP\tzoom in (faster with SHIFT)\n"
          "\tDOWN\tzoom out (faster with SHIFT)\n"
          "\tJ\tJulia mode\n"
          "\tM\tMandelbrot mode\n"
          "\t0\treset inflections\n"
          "\t-\tundo add inflection point\n"
          "\t=\tredo add inflection point\n"
          "\tH\tshow this help\n"
          "mouse controls:\n"
          "\tWHEEL\tzoom about mouse cursor position\n"
          "\tLEFT\tadd inflection point at cursor position\n"
          "\tRIGHT\tundo add inflection point\n"
          "\tMIDDLE\trecenter window about mouse cursor position\n"
        );
        break;
    }
}

vec2 mouse_position(GLFWwindow *window)
{
  double w = S.width;
  double h = S.height;
  double x = 0, y = 0;
  glfwGetCursorPos(window, &x, &y);
  double dx = 2 * ((x + 0.5) / w - 0.5);
  double dy = 2 * (0.5 - (y + 0.5) / h);
  vec2 p = vec2(float(dx), float(dy));
  vec2 t = S.center + cMul(S.radius, S.aspect * p);
  return t;
}

void scrollcb(GLFWwindow* window, double xoffset, double yoffset)
{
  (void) xoffset;
  float g = pow(0.9, yoffset);
  S.center = mix(mouse_position(window), S.center, g);
  S.radius *= g;
  glUniform2fv(S.u_center, 1, &S.center[0]);
  glUniform2fv(S.u_radius, 1, &S.radius[0]);
}

void buttoncb(GLFWwindow *window, int button, int action, int mods)
{
  (void) mods;
  if (action == GLFW_RELEASE)
    switch (button)
    {
      case GLFW_MOUSE_BUTTON_LEFT:
        add_point(mouse_position(window));
        break;
      case GLFW_MOUSE_BUTTON_RIGHT:
        undo_add_point();
        break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
        S.center = mouse_position(window);
        glUniform2fv(S.u_center, 1, &S.center[0]);
        break;
    }
}

int main(int argc, char **argv)
{
  (void) argc;
  (void) argv;

  S.width = 1280;
  S.height = 720;
  S.max_count = 0;

  glfwInit();
  glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
  GLFWwindow *window = glfwCreateWindow(S.width, S.height, "inflector-gadget", 0, 0);
  glfwMakeContextCurrent(window);
  glewExperimental = GL_TRUE;
  glewInit();
  glGetError(); // discard common error from glew 
  glfwSetMouseButtonCallback(window, buttoncb);
  glfwSetScrollCallback(window, scrollcb);
  glfwSetKeyCallback(window, keycb);

  S.center = vec2(0.0f, 0.0f);
  S.radius = vec2(2.0f, 0.0f);
  S.aspect = vec2(S.width * 1.0f / S.height, 1.0f);
  S.julia  = true;
  S.count = 0;
  memset(&S.inflection[0], 0, MAXCOUNT * sizeof(vec2));

  GLuint program = vertex_fragment_shader(render_vert, render_frag);
  glUseProgram(program);
  S.u_center     = glGetUniformLocation(program, "center");
  S.u_radius     = glGetUniformLocation(program, "radius");
  S.u_aspect     = glGetUniformLocation(program, "aspect");
  S.u_julia      = glGetUniformLocation(program, "julia");
  S.u_count      = glGetUniformLocation(program, "count");
  S.u_inflection = glGetUniformLocation(program, "inflection");
  S.a_point      = glGetAttribLocation (program, "point");
  glUniform2fv(S.u_center, 1, &S.center[0]);
  glUniform2fv(S.u_radius, 1, &S.radius[0]);
  glUniform2fv(S.u_aspect, 1, &S.aspect[0]);
  glUniform1i (S.u_julia, S.julia);
  update_inflections();

  GLuint vao = 0;
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);

  float vdata[8] =
    { -1.0, -1.0
    ,  1.0, -1.0
    , -1.0,  1.0
    ,  1.0,  1.0
    };
  GLuint vbo = 0;
  glGenBuffers(1, &vbo);
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  glBufferData(GL_ARRAY_BUFFER, 8 * sizeof(float), &vdata[0], GL_STATIC_DRAW);
  glVertexAttribPointer(S.a_point, 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(S.a_point);

  while (1) {
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glfwSwapBuffers(window);
    glfwWaitEvents();
    if (glfwWindowShouldClose(window))
      break;
  }

  glfwDestroyWindow(window);
  glfwTerminate();
  return 0;
}
